#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <pthread.h>
#include <getopt.h>
#include "bytes.h"
#include "pcp_lib.h"
#include "pcp_options.h"
#include "xmalloc.h"

static void set_cmd_type(pcp_options *opts, int cmd_type)
{
    if (opts->cmd_type != PCP_NULL_CMD) {
        printf("Only one cmd type allowed\n");
        exit(0);
    }
    opts->cmd_type = cmd_type;
}

static int file_exists(const char *filename)
{
#ifdef HAVE_ACCESS
    return access(filename, F_OK) >= 0;
#else
    struct stat buf;
    return stat(filename, &buf) >= 0;
#endif
}

static void print_help()
{
    printf("NOTE: add help\n");
}

void pcp_options_init(pcp_options *opts)
{
    opts->cmd_type     = PCP_NULL_CMD;
    opts->verbose      = 0;
    opts->fanout       = PCP_DEF_FANOUT;
    opts->frag_size    = PCP_DEF_FRAG_SIZE;
    opts->best_effort  = 0;
    opts->args_index   = 1;
}

void pcp_options_read_pcprc(pcp_options *opts)
{
    char  *pcprc, *home;
    char  attr[128];
    int   value;
    FILE  *f;

    if ((home = getenv("HOME")) == NULL)
        return;
    pcprc = (char *)xmalloc(strlen(home) + SLASHBYTE_SIZE + strlen(".pcprc") + 
                            NULLBYTE_SIZE);
    sprintf(pcprc, "%s/.pcprc", home);

    if (!file_exists(pcprc)) {
        xfree(pcprc);
        return;
    }
    if ((f = fopen(pcprc, "r")) == NULL) {
        xfree(pcprc);
        return;
    }
    while (fscanf(f, "%127s %d\n", attr, &value) != EOF) {
        if (strcmp(attr, "fanout") == 0) {
            opts->fanout = value;
            if (opts->fanout < 1) {
                fprintf(stderr, "Fanout %d must be >= 1\n", opts->fanout);
                exit(0);
            }
        }
        else if (strcmp(attr, "frag_size") == 0) {
            opts->frag_size = value;
            if (opts->frag_size < 1) {
                fprintf(stderr, "Fragment size %d must be >= 1\n",
                        opts->frag_size);
                exit(0);
            }
        }
        else {
            fprintf(stderr, "Unrecognized option %s in .pcprc\n", attr);
        }
    }
    fclose(f);
    xfree(pcprc);
}

void pcp_options_parse(pcp_options *opts, int argc, char **argv)
{
    int c;
    extern char *optarg;
    static struct option long_options[] =
    {
        { "best-effort", no_argument, NULL, 'b' },
        { "checksum", no_argument, NULL, 'c' },
        { "delete", no_argument, NULL, 'd' },
        { "help", no_argument, NULL, 'h' },
        { "verbose", no_argument, NULL, 'v' },
        { "fanout", required_argument, NULL, 'f' },
        { "frag-size", required_argument, NULL, 'n' },
        { 0, 0, 0, 0 }
    };

    while ((c = getopt_long (argc, argv, "bcdhvVf:n:", long_options, 
                             (int *)0)) != EOF) {
        switch (c) {
        case 'b':
            opts->best_effort = 1;
            opts->args_index++;
            break;
        case 'c':
            set_cmd_type(opts, PCP_CHECKSUM_CMD);
            opts->args_index++;
            break;
        case 'd':
            set_cmd_type(opts, PCP_DELETE_CMD);
            opts->args_index++;
            break;
        case 'h':
            print_help();
            exit(0);
        case 'v':
            opts->verbose = 1;
            opts->args_index++;
            break;
        case 'f':
            opts->fanout = atoi(optarg);
            if (opts->fanout < 1) {
                fprintf(stderr, "Fanout %d must be >= 1\n", opts->fanout);
                exit(0);
            }
            opts->args_index += 2;
            break;
        case 'n':
            opts->frag_size = atoi(optarg);
            if (opts->frag_size < 1) {
                fprintf(stderr, "Fragment size %d must be >= 1\n",
                        opts->frag_size);
                exit(0);
            }
            opts->args_index += 2;
            break;
        }
    }
    if (opts->cmd_type == PCP_NULL_CMD)
        set_cmd_type(opts, PCP_WRITE_CMD); /* default */
}
