#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "bytes.h"
#include "e_error.h"
#include "hash.h"
#include "msg.h"
#include "net.h"
#include "tree.h"
#include "xmalloc.h"

static void compute_parents(int *parents, int nhosts, int fanout)
{
    int i, j, c;

    parents[0] = -1;
    for (i = 0; i < nhosts; i++) {
        for (c = 1; c <= fanout; c++) {
            j = i * fanout + c;
            if (j >= nhosts)
                return;
            parents[j] = i;
        }
    }
}

static int has_cycle(tree *t)
{
    int        i, klen, cycle = 0; 
    int        present = 1;
    void       *key;
    hash_table *hash;

    /* Checks for duplicates in [1,n-1] (ignores root) */
    hash_table_create(&hash, 191);
    for (i = 1; i < t->nhosts; i++) {
        key  = t->ips[i];
        klen = strlen(t->ips[i]) + NULLBYTE_SIZE;
        if (!hash_table_insert(hash, key, klen, &present, sizeof(int))) {
            cycle = 1;
            break;
        }
    }
    hash_table_destroy(hash);
    return (!cycle) ? PCP_OK : PCP_GRAPH_CYCLE_ERROR;
}

/* Ensures that src has at least len bytes left */
int safe_memcpy(void *dst, void *src, int len, int src_len)
{
    int new_src_len;

    new_src_len = src_len - len;
    if (new_src_len < 0)
        return PCP_INTERNAL_ERROR;
    memcpy(dst, src, len);
    return PCP_OK;
}

int tree_create(tree **t, int nhosts, char **ips, int fanout) 
{
    int i;

    (*t) = (tree *)xmalloc(sizeof(tree));
    (*t)->nhosts  = nhosts;
    (*t)->ips     = (char **)xmalloc(nhosts * sizeof(char *));
    (*t)->parents = (int *)xmalloc(nhosts * sizeof(int));
    (*t)->fanout  = fanout;
    for (i = 0; i < nhosts; i++) {
        (*t)->ips[i] = (char *)xmalloc(IP_STRLEN);
        memcpy((*t)->ips[i], ips[i], IP_STRLEN);
    }
    compute_parents((*t)->parents, (*t)->nhosts, (*t)->fanout);
    if (has_cycle(*t))
        goto cleanup;
    return PCP_OK;

 cleanup:
    for (i = 0; i < nhosts; i++)
        xfree((*t)->ips[i]);
    xfree((*t)->parents);
    xfree((*t)->ips);
    xfree((*t));
    *t = NULL;
    return PCP_GRAPH_CYCLE_ERROR;
}

int tree_create_from_buf(tree **t, void *buf, int buf_len)
{
    int  i, j, len, plen;
    void *p;

    (*t) = (tree *)xmalloc(sizeof(tree));
    p    = buf;
    plen = buf_len;

    /* Copy nhosts */
    len = sizeof((*t)->nhosts);
    if (safe_memcpy(&(*t)->nhosts, p, len, plen) != PCP_OK)
        goto cleanup0;
    p    += len;
    plen -= len;

    /* Copy ips */
    (*t)->ips = (char **)xmalloc((*t)->nhosts * sizeof(char *));
    for (i = 0; i < (*t)->nhosts; i++) {
        (*t)->ips[i] = (char *)xmalloc(IP_STRLEN);
        if (safe_memcpy((*t)->ips[i], p, IP_STRLEN, plen) != PCP_OK)
            goto cleanup1;
        p    += IP_STRLEN;
        plen -= IP_STRLEN;
    }

    /* Copy parents */
    len = (*t)->nhosts * sizeof(int);
    (*t)->parents = (int *)xmalloc(len);    
    if (safe_memcpy((*t)->parents, p, len, plen) != PCP_OK)
        goto cleanup2;
    p    += len;
    plen -= len;

    /* Copy fanout */
    len = sizeof((*t)->fanout);
    if (safe_memcpy(&(*t)->fanout, p, len, plen) != PCP_OK)
        goto cleanup2;
    p    += len;
    plen -= len;

    if (has_cycle(*t))
        goto cleanup2;
    return PCP_OK;

 cleanup2:
    xfree((*t)->parents);
 cleanup1:
    for (j = 0; j < i; j++)
        xfree((*t)->ips[j]);
    xfree((*t)->ips);
 cleanup0:
    xfree((*t));
    *t = NULL;
    return PCP_GRAPH_CYCLE_ERROR;
}

void *tree_malloc_buf(tree *t, int *buf_len)
{
    int  i;
    void *buf, *p;

    *buf_len = sizeof(t->nhosts) + t->nhosts * IP_STRLEN + 
        t->nhosts * sizeof(int) + sizeof(t->fanout);
    p = buf = (void *)xmalloc(*buf_len);
    memcpy(p, &t->nhosts, sizeof(t->nhosts));
    p += sizeof(t->nhosts);
    for (i = 0; i < t->nhosts; i++) {
        memcpy(p, t->ips[i], IP_STRLEN);
        p += IP_STRLEN;
    }
    memcpy(p, &t->parents, t->nhosts * sizeof(int));
    p += t->nhosts * sizeof(int);
    memcpy(p, &t->fanout, sizeof(t->fanout));
    p += sizeof(t->fanout);
    return buf;
}

void tree_destroy(tree *t)
{
    int i;

    for (i = 0; i < t->nhosts; i++) 
        xfree(t->ips[i]);
    xfree(t->ips);
    xfree(t);
}

int tree_parent(tree *t, int index)
{
    return t->parents[index];
}

int tree_child(tree *t, int index, int child_num)
{
    int first, child;

    first = (index * t->fanout) + 1;
    first = (first < t->nhosts) ? first : -1;
    if (first == -1)
        return -1;

    child = first + child_num;
    return (child < t->nhosts) ? child : -1;
}

int tree_num_children(tree *t, int index)
{
    int first, last;
    
    first = (index * t->fanout) + 1;
    first = (first < t->nhosts) ? first : -1;
    if (first == -1)
        return 0;

    last = first + (t->fanout - 1);
    last = (last < t->nhosts) ? last : t->nhosts - 1;
    return (last - first + 1);
}

int tree_node(tree *t, char *host)
{
    int i;
    for (i = 0; i < t->nhosts; i++)
        if (strcmp(t->ips[i], host) == 0)
            return i;
    return -1;
}

/* Ignore root */
int tree_node_nr(tree *t, char *host)
{
    int i;
    for (i = 1; i < t->nhosts; i++)
        if (strcmp(t->ips[i], host) == 0)
            return i;
    return -1;
}
